/*********
  Rui Santos
  Complete project details at https://RandomNerdTutorials.com/esp32-esp8266-input-data-html-form/
  
  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files.
  
  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.
*********/

// include wifi password
#include "config.h"
#include <Arduino.h>
#include <WiFi.h>
#include <AsyncTCP.h>
#include <SPIFFS.h>
#include <Preferences.h>
#include <arduino-timer.h>
#include <jled.h> // jled
#include "jled/morse.h" //jled
#include <Adafruit_BusIO_Register.h> // for DS3231
#include <RTClib.h> // for DS3231
#include <string>

// download zip from https://github.com/me-no-dev/ESPAsyncWebServer and install.
#include <ESPAsyncWebServer.h>

AsyncWebServer server(80);

RTC_DS3231 rtc; // set up RTC
const int alarmPin = 4; // pin to monitor for RTC alarms

// Read from config.h
const char* ssid = WIFI_SSID;
const char* password = WIFI_PASSWORD;

const char* PARAM_STRING = "inputString";
const char* PARAM_SEND = "inputSend";
const char* PARAM_WPM = "inputWPM";
const char* PARAM_MSG = "inputMsg";
const char* PARAM_FLOAT = "inputFloat";
const char* PARAM_TIME = "inputTimeUnix";
const char* PARAM_START = "inputStartTimeUnix";
const char* PARAM_RUNNING = "programRunning";

// Global variables
String yourInputString;
int yourInputSend;
int yourInputWPM;
int yourInputMsg;
int yourInputMsg_old; // to save previous state and check changes
float yourInputFloat;
uint32_t yourInputTime; //to keep time
uint32_t yourInputStartTimeUnix;
int yourProgramRunning;

// HTML web page to handle 3 input fields (inputString, inputSend, inputFloat)
const char index_html[] PROGMEM = R"rawliteral(
<!DOCTYPE HTML><html><head>
  <link rel="icon" href="data:,">
  <title>ESP Input Form</title>
  <meta name="viewport" content="width=device-width, initial-scale=1">
  <script type="text/javascript">
    // Utility from https://webreflection.medium.com/using-the-input-datetime-local-9503e7efdce
    Date.prototype.toDatetimeLocal = function toDatetimeLocal() {
    var
      date = this,
      ten = function (i) {
        return (i < 10 ? '0' : '') + i;
      },
      YYYY = date.getFullYear(),
      MM = ten(date.getMonth() + 1),
      DD = ten(date.getDate()),
      HH = ten(date.getHours()),
      II = ten(date.getMinutes()),
      SS = ten(date.getSeconds())
    ;
    return YYYY + '-' + MM + '-' + DD + 'T' +
             HH + ':' + II + ':' + SS;
    }

    // Submit timestamps as unix seconds when form is submitted
    var putDate = function(form) {
      form.inputTimeUnix.value = Math.floor(Date.now() / 1000);// - new Date().getTimezoneOffset()*60;
      form.inputStartTimeUnix.value = ((Date.parse(js_start_time_unix_entry.value))/1000);
      //document.getElementById("js_start_time_unix").value = ((Date.parse(js_start_time_unix_entry.value))/1000);
    }
    // Fill in page values
    window.onload = function() {
      s = %inputStartTimeUnix%;
      current_start = new Date(s * 1000);
      document.getElementById('current-start').innerHTML = current_start.toLocaleString();
      // Show the local time as a string
      local_time_unix = new Date().toLocaleString();//toUTCString();
      document.getElementById('local-time-unix').innerHTML = local_time_unix.toString();
      // Fill in the start time field as local time
      document.getElementById('js_start_time_unix_entry').value = current_start.toDatetimeLocal();

      // Fill in the other form fields
      document.getElementById("send-program").value = %inputSend%;
      document.getElementById("message").value = %inputMsg%;
    }

  </script></head><body>

  <p>Local time: <b><span id=local-time-unix></span></b></p>

  <form action="/get"  onsubmit="putDate(this);" accept-charset=utf-8>
    
    <p>Sending program (cycle doesn't work yet) (current value: <b>%inputSend%</b>):
    <select name="inputSend" id="send-program">
      <option value="0" >0 - Off</option>
      <option value="1">1 - Continuous</option>
      <option value="2">2 - Cycle</option>
    </select><br>

    Message (current value <b>%inputMsg%</b>):
    <select name="inputMsg" id="message">
      <option value="0">0 - TEST TEST TEST DE W1CDN</option>
      <option value="1">1 - MOE</option>
      <option value="2">2 - MOI</option>
      <option value="3">3 - MOS</option>
      <option value="4">4 - MOH</option>
      <option value="5">5 - MO5</option>
    </select></p>

    <p>Cycle start time <input type="datetime-local" id="js_start_time_unix_entry" /><br>
    Current value: <b><span id=current-start></span></b>, but only day-of-month HH:MM:SS are used.
    Only applies when <em>Sending Program</em> is set to "2 - Cycle". 
    You cannot set a cycle start date more than a month in advance.
    <!-- JS converts the entered start time to a unix timestamp, and copies that value
    to this hidden field so the user doesn't have to see it. -->
    <input type="hidden" name="inputStartTimeUnix" id="js_start_time_unix" /></p>
    
    <!-- This field is hidden so people don't change the submit time (it will be wrong).
    The value is automatically filled in with JS. -->
    <input type="hidden" name="inputTimeUnix" id="js_time_unix">

    <!-- Extra fields just in case I need them -->  
    <input type="hidden" name="inputWPM" value = %inputWPM%>
    <input type="hidden" name="inputString" value = %inputString%>
    <input type="hidden" name="inputFloat" value = %inputFloat%>

    <input type="submit" value="Submit"">
  </form>
  <iframe style="display:none" name="hidden-form" id="hidden-form"></iframe>
  <script type="text/javascript">
    
  </script>
</body></html>)rawliteral";

// Assign output variables to GPIO pins
const int keyer = 32; //LED_BUILTIN for on-board (dev);//26 for LED; //32 for transmitter keyer
const int blinker = LED_BUILTIN;

// Timers
//auto timer = timer_create_default();
Timer<1> timer;

// Example from https://github.com/contrem/arduino-timer#examples
bool toggle_led(void *) {
  //Serial.print("Timer time: ");
  //Serial.println(rtc.now().timestamp());
  digitalWrite(LED_BUILTIN, !digitalRead(LED_BUILTIN)); // toggle the LED
  return true; // keep timer active? true
}

void notFound(AsyncWebServerRequest *request) {
  request->send(404, "text/plain", "Not found");
}

String readFile(fs::FS &fs, const char * path){
  //Serial.printf("Reading file: %s\r\n", path);
  File file = fs.open(path, "r");
  if(!file || file.isDirectory()){
    Serial.println("- empty file or failed to open file");
    return String();
  }
  //Serial.println("- read from file:");
  String fileContent;
  while(file.available()){
    fileContent+=String((char)file.read());
  }
  file.close();
  //Serial.println(fileContent);
  return fileContent;
}

void writeFile(fs::FS &fs, const char * path, const char * message){
  Serial.printf("Writing file: %s\r\n", path);
  File file = fs.open(path, "w");
  if(!file){
    Serial.println("- failed to open file for writing");
    return;
  }
  if(file.print(message)){
    Serial.println("- file written");
  } else {
    Serial.println("- write failed");
  }
  file.close();
}

// Replaces placeholder in web UI with stored values
String processor(const String& var){
  //Serial.println(var);
  if(var == "inputString"){
    return readFile(SPIFFS, "/inputString.txt");
  }
  else if(var == "inputSend"){
    return readFile(SPIFFS, "/inputSend.txt");
  }
  else if(var == "inputWPM"){
    return readFile(SPIFFS, "/inputWPM.txt");
  }
  else if(var == "inputMsg"){
    return readFile(SPIFFS, "/inputMsg.txt");
  }
  else if(var == "inputFloat"){
    return readFile(SPIFFS, "/inputFloat.txt");
  } else if(var == "inputStartTimeUnix"){
    // Webform breaks if this value is empty.
    String temp = readFile(SPIFFS, "/inputStartTimeUnix.txt");
    if(temp == ""){
      temp = "0";
    }
    return temp;
  } 
  return String();
}

//jled from https://github.com/jandelgado/jled/blob/master/examples/morse/morse_effect.h
class MorseEffect : public jled::BrightnessEvaluator {
    Morse morse_;
    // duration of a single 'dit' in ms
    const uint16_t speed_;

 public:
    explicit MorseEffect(const char* message, uint16_t speed = 200)
        : morse_(message), speed_(speed) {}

    uint8_t Eval(uint32_t t) const override {
        const auto pos = t / speed_;
        if (pos >= morse_.size()) return 0;
        return morse_.test(pos) ? 255 : 0;
    }

    uint16_t Period() const override { return (morse_.size() + 1) * speed_; }
};

// Speed is milliseconds per dit, which is 1000 * (60 / (50 * WPM))
// 60 is 20 wpm, 120 is 10 wpm, 90 is 15 wpm, etc.
// https://morsecode.world/international/timing.html
float wpm = 10;
float ms_per_dit = 1000 * (60 / (50 * wpm));
int word_space_ms = ms_per_dit * 7;
// Hardcoding messages and WPM  for now, will come back and make it more flexible.
MorseEffect morseEffectTEST("TEST TEST TEST DE W1CDN", ms_per_dit);
MorseEffect morseEffectMOE("MOE", ms_per_dit);
MorseEffect morseEffectMOI("MOI", ms_per_dit);
MorseEffect morseEffectMOS("MOS", ms_per_dit);
MorseEffect morseEffectMOH("MOH", ms_per_dit);
MorseEffect morseEffectMO5("MO5", ms_per_dit);

// CW for keyer
auto morseTEST =
    JLed(keyer).UserFunc(&morseEffectTEST).DelayAfter(word_space_ms).Forever();
auto morseMOE =
    JLed(keyer).UserFunc(&morseEffectMOE).DelayAfter(word_space_ms).Forever();
auto morseMOI =
    JLed(keyer).UserFunc(&morseEffectMOI).DelayAfter(word_space_ms).Forever();
auto morseMOS =
    JLed(keyer).UserFunc(&morseEffectMOS).DelayAfter(word_space_ms).Forever();
auto morseMOH =
    JLed(keyer).UserFunc(&morseEffectMOH).DelayAfter(word_space_ms).Forever();
auto morseMO5 =
    JLed(keyer).UserFunc(&morseEffectMO5).DelayAfter(word_space_ms).Forever();
auto morseToSend = morseTEST; // set this up to overwrite later

// CW for blinker
auto morseTEST_blink =
    JLed(blinker).UserFunc(&morseEffectTEST).DelayAfter(word_space_ms).Forever();
auto morseMOE_blink =
    JLed(blinker).UserFunc(&morseEffectMOE).DelayAfter(word_space_ms).Forever();
auto morseMOI_blink =
    JLed(blinker).UserFunc(&morseEffectMOI).DelayAfter(word_space_ms).Forever();
auto morseMOS_blink =
    JLed(blinker).UserFunc(&morseEffectMOS).DelayAfter(word_space_ms).Forever();
auto morseMOH_blink =
    JLed(blinker).UserFunc(&morseEffectMOH).DelayAfter(word_space_ms).Forever();
auto morseMO5_blink =
    JLed(blinker).UserFunc(&morseEffectMO5).DelayAfter(word_space_ms).Forever();
auto morseToSend_blink = morseTEST_blink; // set this up to overwrite later

//================================================================================
// start_program(): a function to start the planned program at the planned time
//================================================================================
bool start_program(){
  Serial.println("The scheduled program has started.");
  yourProgramRunning = 1;
  return false;
}

//================================================================================
// setup(): stuff that only gets done once, after power up (KB1OIQ's description)
//================================================================================
void setup() {
  Serial.begin(115200);

  Serial.print("morseEffectTEST.Period(): ");
  Serial.println(morseEffectTEST.Period());

  pinMode(alarmPin, INPUT_PULLUP); // Set alarm pin as pullup

  if (! rtc.begin()) {
    Serial.println("Couldn't find RTC");
    Serial.flush();
    while (1) delay(10);
  }

  if (rtc.lostPower()) {
    Serial.println("RTC lost power, let's set the time!");
    // When time needs to be set on a new device, or after a power loss, the
    // following line sets the RTC to the date & time this sketch was compiled
    rtc.adjust(DateTime(__DATE__, __TIME__));
    // This line sets the RTC with an explicit date & time, for example to set
    // January 21, 2014 at 3am you would call:
    //rtc.adjust(DateTime(2023, 9, 2, 17, 32, 0));
  }

  // Are there any RTC alarms set?
  DateTime alarm_one = rtc.getAlarm1(); // Get the current time
  char buff[] = "Alarm 1 set for at hh:mm:ss DDD, DD MMM YYYY";
  Serial.print(alarm_one.toString(buff));
  Serial.println(" (only HH:MM:SS day-of-month are accurate)");

  // Initialize the output variables as outputs
  pinMode(keyer, OUTPUT);
  pinMode(blinker, OUTPUT);
  // Set outputs to LOW
  digitalWrite(keyer, LOW);
  digitalWrite(blinker, LOW);

  // Initialize SPIFFS
  SPIFFS.begin(true);
    if(!SPIFFS.begin(true)){
      Serial.println("An Error has occurred while mounting SPIFFS");
      return;
    }
    if(!SPIFFS.begin()){
      Serial.println("An Error has occurred while mounting SPIFFS");
      return;
    }



  // Read in existing data
  yourInputString = readFile(SPIFFS, "/inputString.txt");
  yourInputSend = readFile(SPIFFS, "/inputSend.txt").toInt();
  yourInputWPM = readFile(SPIFFS, "/inputWPM.txt").toInt();
  yourInputMsg = readFile(SPIFFS, "/inputMsg.txt").toInt();
  yourInputFloat = readFile(SPIFFS, "/inputFloat.txt").toFloat();
  yourInputStartTimeUnix = readFile(SPIFFS, "/inputStartTimeUnix.txt").toInt();

  // On restart, keep doing what you were doing before
  yourInputMsg_old = yourInputMsg;
  if(yourInputMsg == 0){
    morseToSend = morseTEST;
    morseToSend_blink = morseTEST_blink;
  } else if(yourInputMsg == 1){
    morseToSend = morseMOE;
    morseToSend_blink = morseMOE_blink;
  } else if(yourInputMsg == 2){
    morseToSend = morseMOI;
    morseToSend_blink = morseMOI_blink;
  } else if(yourInputMsg == 3){
    morseToSend = morseMOS;
    morseToSend_blink = morseMOS_blink;
  } else if(yourInputMsg == 4){
    morseToSend = morseMOH;
    morseToSend_blink = morseMOH_blink;
  } else if(yourInputMsg == 5){
    morseToSend = morseMO5;
    morseToSend_blink = morseMO5_blink;
  }

  WiFi.mode(WIFI_STA);
  WiFi.begin(ssid, password);
  if (WiFi.waitForConnectResult() != WL_CONNECTED) {
    Serial.println("WiFi Failed!");
    return;
  }
  Serial.println();
  Serial.print("IP Address: ");
  Serial.println(WiFi.localIP());

  // Send web page with input fields to client
  server.on("/", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send_P(200, "text/html", index_html, processor);
  });

  // Send a GET request to <ESP_IP>/get?inputString=<inputMessage>
  server.on("/get", HTTP_GET, [] (AsyncWebServerRequest *request) {
    String inputMessage;
    // GET inputString value on <ESP_IP>/get?inputString=<inputMessage>
    if (request->hasParam(PARAM_STRING)) {
      inputMessage = request->getParam(PARAM_STRING)->value();
      writeFile(SPIFFS, "/inputString.txt", inputMessage.c_str());
      yourInputString = inputMessage;
    }
    // GET inputSend value on <ESP_IP>/get?inputSend=<inputMessage>
    if (request->hasParam(PARAM_SEND)) {
      inputMessage = request->getParam(PARAM_SEND)->value();
      writeFile(SPIFFS, "/inputSend.txt", inputMessage.c_str());
      yourInputSend = inputMessage.toInt();
      // if not running a program, set the program running off
      if(yourInputSend != 2){
        yourProgramRunning = 0;
      }
    }
    // GET inputWPM value on <ESP_IP>/get?inputWPM=<inputMessage>
    if (request->hasParam(PARAM_WPM)) {
      inputMessage = request->getParam(PARAM_WPM)->value();
      writeFile(SPIFFS, "/inputWPM.txt", inputMessage.c_str());
      yourInputWPM = inputMessage.toInt();
    }
    // GET inputMsg value on <ESP_IP>/get?inputMsg=<inputMessage>
    if (request->hasParam(PARAM_MSG)) {
      inputMessage = request->getParam(PARAM_MSG)->value();
      writeFile(SPIFFS, "/inputMsg.txt", inputMessage.c_str());
      // save previous state
      yourInputMsg_old = yourInputMsg;
      yourInputMsg = inputMessage.toInt();
    }
    // GET inputTimeUnix value on <ESP_IP>/get?inputTimeUnix=<inputMessage>
    if (request->hasParam(PARAM_TIME)) {
      inputMessage = request->getParam(PARAM_TIME)->value();
      //https://stackoverflow.com/a/22733127/2152245
      yourInputTime = atol(inputMessage.c_str());
      Serial.print("yourInputTime: ");
      Serial.println(yourInputTime);
      // update the RTC time
      rtc.adjust(DateTime(yourInputTime));
      DateTime now = rtc.now();
      // Might work to fix random errors? If date is far in the future,
      // try to update again.
      // replace if with while if you want it to try a bunch...
      if(now.year() > 2040){
        Serial.print("Year is ");
        Serial.println(now.year());
        Serial.println("RTC can't set time. Trying again.");
        rtc.adjust(DateTime(yourInputTime));
      }

      Serial.print("UTC time from browser: ");
      Serial.print(now.year(), DEC);
      Serial.print('/');
      Serial.print(now.month(), DEC);
      Serial.print('/');
      Serial.print(now.day(), DEC);
      Serial.print(" (");
      Serial.print(now.dayOfTheWeek());
      Serial.print(") ");
      Serial.print(now.hour(), DEC);
      Serial.print(':');
      Serial.print(now.minute(), DEC);
      Serial.print(':');
      Serial.print(now.second(), DEC);
      Serial.println();

      Serial.print("rtc.now().unixtime(): ");
      Serial.println(rtc.now().unixtime());
    }
    // GET inputFloat value on <ESP_IP>/get?inputFloat=<inputMessage>
    if (request->hasParam(PARAM_FLOAT)) {
      inputMessage = request->getParam(PARAM_FLOAT)->value();
      writeFile(SPIFFS, "/inputFloat.txt", inputMessage.c_str());
      yourInputFloat = inputMessage.toFloat();
    }
    // GET inputStartTimeUnix value on <ESP_IP>/get?inputStartTimeUnix=<inputMessage>
    if (request->hasParam(PARAM_START)) {
      inputMessage = request->getParam(PARAM_START)->value();
      Serial.println(inputMessage);
      // if a start time isn't entered, don't overwrite the old one
      //if(!(inputMessage != NULL && inputMessage[0] == '\0')){
        writeFile(SPIFFS, "/inputStartTimeUnix.txt", inputMessage.c_str());
        yourInputStartTimeUnix = atol(inputMessage.c_str());
      //}
      Serial.println(yourInputStartTimeUnix);

      // We can't use arduino-timer for starting a program because 
      // it relies on millis(), which reset on power cycle.
      // timer.at(millis() + 10000, toggle_led);
      // Serial.println(millis());
      // auto active_tasks = timer.size();
      // Serial.println(active_tasks);
      
      // Use alarm built into RTC
      rtc.setAlarm1(DateTime(yourInputStartTimeUnix), DS3231_A1_Date);
      //rtc.setAlarm1(DateTime(2020, 6, 25, 15, 34, 0), DS3231_A2_Date);
      DateTime alarm_one = rtc.getAlarm1(); // Get the current alarm time
      char buff[] = "Alarm 1 set for at hh:mm:ss DDD, DD MMM YYYY";
      Serial.print(alarm_one.toString(buff));
      Serial.println(" (only HH:MM:SS day-of-month are accurate)");
    }
    // https://techtutorialsx.com/2018/01/14/esp32-arduino-http-server-external-and-internal-redirects/
    request->redirect("/");
  });
  server.onNotFound(notFound);
  server.begin();

}

void loop() {
    // Timers
  timer.tick();

  // This statement from https://github.com/garrysblog/DS3231-Alarm-With-Adafruit-RTClib-Library/blob/master/DS3231-RTClib-Adafruit-Alarm-Poll-alarmFired/DS3231-RTClib-Adafruit-Alarm-Poll-alarmFired.ino
  // Check if alarm by polling SQW alarm pin
  if (yourInputSend == 2 & digitalRead(alarmPin) == LOW) {
    // Print current time and date
    DateTime now = rtc.now(); // Get the current time
    char buff[] = "Alarm triggered at hh:mm:ss DDD, DD MMM YYYY";
    Serial.println(now.toString(buff));

    // Disable and clear alarm
    rtc.clearAlarm(1); 
    rtc.clearAlarm(2); // clear the other one just in case 
  }

  // See which message we are sending
  // Only do this when the message has been updated.
  if(yourInputMsg != yourInputMsg_old){
    //morseToSend.Stop(JLed::eStopMode::FULL_OFF).Update();
    if(yourInputMsg == 0){
    morseToSend = morseTEST;
    morseToSend_blink = morseTEST_blink;
    // Experiment with timing
    int step_length = 30000; // 30 secs
    int repeats = step_length / morseEffectTEST.Period();
    morseToSend_blink.UserFunc(&morseEffectTEST).DelayAfter(word_space_ms).Repeat(repeats);
  
  } else if(yourInputMsg == 1){
    morseToSend = morseMOE;
    morseToSend_blink = morseMOE_blink;
  } else if(yourInputMsg == 2){
    morseToSend = morseMOI;
    morseToSend_blink = morseMOI_blink;
  } else if(yourInputMsg == 3){
    morseToSend = morseMOS;
    morseToSend_blink = morseMOS_blink;
  } else if(yourInputMsg == 4){
    morseToSend = morseMOH;
    morseToSend_blink = morseMOH_blink;
  } else if(yourInputMsg == 5){
    morseToSend = morseMO5;
    morseToSend_blink = morseMO5_blink;
  }
    // Keeps the key from locking up
    yourInputMsg_old = yourInputMsg;
  }


  // if you want to send continuous code, and it's not sending, then start it up
  if((yourInputSend == 1) & (morseToSend.IsRunning() == false)){
    //jled
    morseToSend.Reset().Update();
    morseToSend_blink.Reset().Update();

  // if you want to send continuous code, and it is sending, keep sending
  } else if((yourInputSend == 1) & (morseToSend.IsRunning() == true)){
    morseToSend.Update();
    morseToSend_blink.Update();
  
  // if you want to send cycle code and it is sending, keep sending
  } else if((yourInputSend == 2) & (morseToSend.IsRunning() == true)){
    morseToSend.Update();
    morseToSend_blink.Update();

  // if you want to send cycle code and it's not sending, then start it up
  } else if((yourInputSend == 2) & (morseToSend.IsRunning() == false)){
    morseToSend.Reset().Update();
    morseToSend_blink.Reset().Update();
      
  // if you don't want to send code
  } else {
    // stop sending and make sure the pin is off
    morseToSend.Stop(JLed::eStopMode::FULL_OFF).Update();
    morseToSend_blink.Stop(JLed::eStopMode::FULL_OFF).Update();
  }
  //morseToSend.Update();

}